# SPDX-License-Identifier: Apache-2.0.
from openlineage.client.run import RunEvent


"""
To implement custom Transport implement Config and Transport classes.

Transport needs to
 * specify class variable `config` that will point to Config class that Transport requires
 * __init__ that will accept specified Config class instance
 * implement `emit` method that will accept RunEvent

Config file is read and parameters there are passed to `from_dict` classmethod.
The config class can have more complex attributes, but needs to be able to
instantiate them in `from_dict` method.

DefaultTransportFactory instantiates custom transports by looking at `type` field in
class config.
"""


class Config:
    @classmethod
    def from_dict(cls, params: dict):
        return cls()


class Transport:
    kind = None
    config = Config

    def emit(self, event: RunEvent):
        raise NotImplementedError()


class TransportFactory:
    def create(self) -> Transport:
        pass



# SPDX-License-Identifier: Apache-2.0.

from typing import Dict, List, Optional
from enum import Enum

import uuid
import attr

from openlineage.client.facet import NominalTimeRunFacet, ParentRunFacet


class RunState(Enum):
    START = 'START'
    COMPLETE = 'COMPLETE'
    ABORT = 'ABORT'
    FAIL = 'FAIL'
    OTHER = 'OTHER'


_RUN_FACETS = [
    NominalTimeRunFacet,
    ParentRunFacet
]


@attr.s
class Run:
    runId: str = attr.ib()
    facets: Dict = attr.ib(factory=dict)

    @runId.validator
    def check(self, attribute, value):
        uuid.UUID(value)


@attr.s
class Job:
    namespace: str = attr.ib()
    name: str = attr.ib()
    facets: Dict = attr.ib(factory=dict)


@attr.s
class Dataset:
    namespace: str = attr.ib()
    name: str = attr.ib()
    facets: Dict = attr.ib(factory=dict)


@attr.s
class InputDataset(Dataset):
    inputFacets: Dict = attr.ib(factory=dict)


@attr.s
class OutputDataset(Dataset):
    outputFacets: Dict = attr.ib(factory=dict)


@attr.s
class RunEvent:
    eventType: RunState = attr.ib(validator=attr.validators.in_(RunState))
    eventTime: str = attr.ib()  # TODO: validate dates
    run: Run = attr.ib()
    job: Job = attr.ib()
    producer: str = attr.ib()
    inputs: Optional[List[Dataset]] = attr.ib(factory=list)
    outputs: Optional[List[Dataset]] = attr.ib(factory=list)



# SPDX-License-Identifier: Apache-2.0.

import json
from enum import Enum
from typing import List, Dict

import attr

try:
    import numpy
except ImportError:
    numpy = None


class Serde:
    @classmethod
    def remove_nulls_and_enums(cls, obj):
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, Dict):
            return dict(filter(
                lambda x: x[1] is not None,
                {k: cls.remove_nulls_and_enums(v) for k, v in obj.items()}.items()
            ))
        if isinstance(obj, List):
            return list(filter(lambda x: x is not None and x != {}, [
                cls.remove_nulls_and_enums(v) for v in obj if v is not None
            ]))

        # Pandas can use numpy.int64 object
        if numpy and isinstance(obj, numpy.int64):
            return int(obj)
        return obj

    @classmethod
    def to_dict(cls, obj):
        if not isinstance(obj, dict):
            obj = attr.asdict(obj)
        return cls.remove_nulls_and_enums(obj)

    @classmethod
    def to_json(cls, obj):
        return json.dumps(
            cls.to_dict(obj),
            sort_keys=True,
            default=lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
        )



# SPDX-License-Identifier: Apache-2.0.
import inspect
from typing import Type

import attr
import importlib
from warnings import warn


def import_from_string(path: str):
    try:
        module_path, target = path.rsplit('.', 1)
        module = importlib.import_module(module_path)
        return getattr(module, target)
    except Exception as e:
        raise ImportError(f"Failed to import {path}") from e


def try_import_from_string(path: str):
    try:
        return import_from_string(path)
    except ImportError as e:
        warn(e.msg)
        return None


def try_import_subclass_from_string(path: str, clazz: Type):
    subclass = try_import_from_string(path)
    if not inspect.isclass(subclass) or not issubclass(subclass, clazz):
        raise TypeError(
            f"Import path {path} - {str(subclass)} has to be class, and subclass of {str(clazz)}"
        )
    return subclass


# Filter dictionary to get only those key: value pairs that have
# key specified in passed attr class
def get_only_specified_fields(clazz, params: dict) -> dict:
    field_keys = [item.name for item in attr.fields(clazz)]
    return {
        key: value for key, value in params.items() if key in field_keys
    }



# SPDX-License-Identifier: Apache-2.0.
import logging

from openlineage.client.run import RunEvent
from openlineage.client.serde import Serde
from openlineage.client.transport.transport import Transport, Config


class ConsoleConfig(Config):
    pass


class ConsoleTransport(Transport):
    kind = 'console'
    config = ConsoleConfig

    def __init__(self, config: ConsoleConfig):
        self.log = logging.getLogger(__name__)

    def emit(self, event: RunEvent):
        self.log.info(Serde.to_json(event))



# SPDX-License-Identifier: Apache-2.0.

import logging
from typing import Optional

import attr
from requests import Session
from requests.adapters import HTTPAdapter

from openlineage.client.run import RunEvent
from openlineage.client.transport import Transport, get_default_factory
from openlineage.client.transport.http import HttpTransport, HttpConfig


@attr.s
class OpenLineageClientOptions:
    timeout: float = attr.ib(default=5.0)
    verify: bool = attr.ib(default=True)
    api_key: str = attr.ib(default=None)
    adapter: HTTPAdapter = attr.ib(default=None)


log = logging.getLogger(__name__)


class OpenLineageClient:
    def __init__(
        self,
        url: Optional[str] = None,
        options: Optional[OpenLineageClientOptions] = None,
        session: Optional[Session] = None,
        transport: Optional[Transport] = None,
    ):
        if url:
            # Backwards compatibility: if URL or options is set, use old path to initialize
            # HTTP transport.
            if not options:
                options = OpenLineageClientOptions()
            if not session:
                session = Session()
            self._initialize_url(url, options, session)
        elif transport:
            self.transport = transport
        else:
            self.transport = get_default_factory().create()

    def _initialize_url(
        self,
        url: str,
        options: OpenLineageClientOptions,
        session: Session
    ):
        self.transport = HttpTransport(HttpConfig.from_options(
            url=url,
            options=options,
            session=session
        ))

    def emit(self, event: RunEvent):
        if not isinstance(event, RunEvent):
            raise ValueError("`emit` only accepts RunEvent class")
        if not self.transport:
            log.error("Tried to emit OpenLineage event, but transport is not configured.")
        else:
            self.transport.emit(event)

    @classmethod
    def from_environment(cls):
        return cls(transport=get_default_factory().create())



# flake8: noqa
from typing import Type

from openlineage.client.transport.transport import Transport, Config, TransportFactory
from openlineage.client.transport.factory import DefaultTransportFactory
from openlineage.client.transport.http import HttpTransport, HttpConfig
from openlineage.client.transport.kafka import KafkaTransport, KafkaConfig
from openlineage.client.transport.console import ConsoleTransport


_factory = DefaultTransportFactory()
_factory.register_transport(HttpTransport.kind, HttpTransport)
_factory.register_transport(KafkaTransport.kind, KafkaTransport)
_factory.register_transport(ConsoleTransport.kind, ConsoleTransport)


def get_default_factory():
    return _factory


# decorator to wrap transports with
def register_transport(clazz: Type[Transport]):
    _factory.register_transport(clazz.kind, clazz)
    return clazz



# SPDX-License-Identifier: Apache-2.0.
from typing import Dict

import attr

from openlineage.client.run import RunEvent
from openlineage.client.serde import Serde
from openlineage.client.transport.transport import Transport, Config
from openlineage.client.utils import get_only_specified_fields


@attr.s
class KafkaConfig(Config):
    # Kafka producer config
    # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    config: Dict[str, str] = attr.ib()

    # Topic on which we should send messages
    topic: str = attr.ib()

    # Set to true if Kafka should flush after each event. The process that emits can be killed in
    # some cases - for example in Airflow integration, so flushing is desirable there.
    flush: bool = attr.ib(default=True)

    @classmethod
    def from_dict(cls, params: dict):
        if 'config' not in params:
            raise RuntimeError("kafka `config` not passed to KafkaConfig")
        if not isinstance(params['config'], dict):
            raise RuntimeError("`config` passed to KafkaConfig must be dict")
        return cls(**get_only_specified_fields(cls, params))


# Very basic transport impl
class KafkaTransport(Transport):
    kind = "kafka"
    config = KafkaConfig

    def __init__(self, config: KafkaConfig):
        import confluent_kafka as kafka
        self.producer = kafka.Producer(config.config)
        self.topic = config.topic
        self.flush = config.flush

    def emit(self, event: RunEvent):
        self.producer.produce(topic=self.topic, value=Serde.to_json(event).encode('utf-8'))
        if self.flush:
            self.producer.flush(timeout=2)



# SPDX-License-Identifier: Apache-2.0.
import logging
from urllib.parse import urljoin

import attr

from typing import Optional, Dict

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import parse_url

from openlineage.client.run import RunEvent
from openlineage.client.serde import Serde
from openlineage.client.transport.transport import Config, Transport
from openlineage.client.utils import get_only_specified_fields, try_import_subclass_from_string

log = logging.getLogger(__name__)


class TokenProvider:
    def __init__(self, config: Dict):
        pass

    def get_bearer(self) -> Optional[str]:
        return None


class ApiKeyTokenProvider(TokenProvider):
    def __init__(self, config: Dict):
        self.api_key = config['api_key']

    def get_bearer(self) -> Optional[str]:
        return f"Bearer {self.api_key}"


def create_token_provider(auth: Dict) -> TokenProvider:
    if 'type' in auth:
        if auth['type'] == 'api_key':
            return ApiKeyTokenProvider(auth)
        try:
            clazz = try_import_subclass_from_string(auth['type'], TokenProvider)
            return clazz(auth)
        except TypeError:
            pass  # already logged
    return TokenProvider({})


@attr.s
class HttpConfig(Config):
    url: str = attr.ib()
    timeout: float = attr.ib(default=5.0)
    # check TLS certificates
    verify: bool = attr.ib(default=True)
    auth: TokenProvider = attr.ib(factory=lambda: TokenProvider({}))
    # not set by TransportFactory
    session: Optional[Session] = attr.ib(factory=Session)
    # not set by TransportFactory
    adapter: Optional[HTTPAdapter] = attr.ib(default=None)

    @classmethod
    def from_dict(cls, params: dict) -> 'HttpConfig':
        if 'url' not in params:
            raise RuntimeError("`url` key not passed to HttpConfig")
        specified_dict = get_only_specified_fields(cls, params)
        specified_dict['auth'] = create_token_provider(specified_dict.get('auth', {}))
        return cls(**specified_dict)

    @classmethod
    def from_options(cls, url: str, options, session: Optional[Session]) -> 'HttpConfig':
        return cls(
            url=url,
            timeout=options.timeout,
            verify=options.verify,
            auth=ApiKeyTokenProvider({"api_key": options.api_key})
            if options.api_key else TokenProvider({}),
            session=session if session else Session(),
            adapter=options.adapter
        )


class HttpTransport(Transport):
    kind = "http"
    config = HttpConfig

    def __init__(self, config: HttpConfig):
        url = config.url.strip()
        try:
            parsed = parse_url(url)
            if not (parsed.scheme and parsed.netloc):
                raise ValueError(f"Need valid url for OpenLineageClient, passed {url}")
        except Exception as e:
            raise ValueError(f"Need valid url for OpenLineageClient, passed {url}. Exception: {e}")
        self.url = url
        self.session = config.session
        self.session.headers['Content-Type'] = 'application/json'
        self.timeout = config.timeout
        self.verify = config.verify

        self._add_auth(config.auth)
        if config.adapter:
            self.set_adapter(config.adapter)

    def set_adapter(self, adapter: HTTPAdapter):
        self.session.mount(self.url, adapter)

    def emit(self, event: RunEvent):
        event = Serde.to_json(event)
        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"Sending openlineage event {event}")
        resp = self.session.post(
            urljoin(self.url, 'api/v1/lineage'),
            event,
            timeout=self.timeout,
            verify=self.verify
        )
        resp.raise_for_status()
        return resp

    def _add_auth(self, token_provider: TokenProvider):
        self.session.headers.update({
            "Authorization": token_provider.get_bearer()
        })



# SPDX-License-Identifier: Apache-2.0.

__version__ = "0.8.2"

DEFAULT_TIMEOUT_MS = 5000
DEFAULT_NAMESPACE_NAME = 'default'
DEFAULT_OPENLINEAGE_URL = 'http://localhost:5000'
DEFAULT_PRODUCER = f"https://github.com/OpenLineage/OpenLineage/tree/{__version__}/client/python"



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0.
#
# -*- coding: utf-8 -*-

from setuptools import setup, find_namespace_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "attrs>=19.3.0",
    "requests>=2.20.0",
    "pyyaml>=5.1.0"
]

extras_require = {
    "tests": ["pytest", "pytest-cov", "mock", "flake8", "requests", "pyyaml"],
    "kafka": ["confluent-kafka"],
}
extras_require["dev"] = set(sum(extras_require.values(), []))

setup(
    name="openlineage-python",
    version="0.8.2",
    description="OpenLineage Python Client",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="OpenLineage",
    packages=find_namespace_packages(include=['openlineage.*']),
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires=">=3.7",
    zip_safe=False,
    keywords="openlineage",
)



# SPDX-License-Identifier: Apache-2.0.
import inspect
import os
import logging
from typing import Type, Union, Optional

from openlineage.client.transport.transport import Config, Transport, TransportFactory
from openlineage.client.utils import try_import_from_string


log = logging.getLogger(__name__)


try:
    import yaml
except ImportError:
    yaml = False


class DefaultTransportFactory(TransportFactory):
    def __init__(self):
        self.transports = {}

    def register_transport(self, type: str, clazz: Union[Type[Transport], str]):
        self.transports[type] = clazz

    def create(self) -> Transport:
        if yaml:
            yml_config = self._try_config_from_yaml()
            if yml_config:
                return self._create_transport(yml_config)
        # Fallback to setting HTTP transport from env variables
        http = self._try_http_from_env_config()
        if http:
            return http
        # If there is no HTTP transport, log events to console
        from openlineage.client.transport.console import ConsoleTransport, ConsoleConfig
        log.warning("Couldn't initialize transport; will print events to console.")
        return ConsoleTransport(ConsoleConfig())

    def _create_transport(self, config: dict):
        transport_type = config['type']

        if transport_type in self.transports:
            transport_class = self.transports[transport_type]
        else:
            transport_class = transport_type

        if isinstance(transport_class, str):
            transport_class = try_import_from_string(transport_class)
        if not inspect.isclass(transport_class) or not issubclass(transport_class, Transport):
            raise TypeError(
                f"Transport {transport_class} has to be class, and subclass of Transport"
            )

        config_class = transport_class.config

        if isinstance(config_class, str):
            config_class = try_import_from_string(config_class)
        if not inspect.isclass(config_class) or not issubclass(config_class, Config):
            raise TypeError(f"Config {config_class} has to be class, and subclass of Config")

        return transport_class(config_class.from_dict(config))

    def _try_config_from_yaml(self) -> Optional[dict]:
        file = self._find_yaml()
        if file:
            try:
                with open(file, 'r') as f:
                    config = yaml.safe_load(f)
                    return config['transport']
            except Exception:
                # Just move to read env vars
                pass
        return

    @staticmethod
    def _find_yaml() -> Optional[str]:
        # Check OPENLINEAGE_CONFIG env variable
        path = os.getenv('OPENLINEAGE_CONFIG', None)
        try:
            if path and os.path.isfile(path) and os.access(path, os.R_OK):
                return path
        except Exception:
            if path:
                log.exception(f"Couldn't read file {path}: ")
            else:
                # We can get different errors depending on system
                pass

        # Check current working directory:
        try:
            cwd = os.getcwd()
            if 'openlineage.yml' in os.listdir(cwd):
                return os.path.join(cwd, 'openlineage.yml')
        except Exception:
            # We can get different errors depending on system
            pass

        # Check $HOME/.openlineage dir
        try:
            path = os.path.expanduser("~/.openlineage")
            if 'openlineage.yml' in os.listdir(path):
                return os.path.join(path, 'openlineage.yml')
        except Exception:
            # We can get different errors depending on system
            pass

    @staticmethod
    def _try_http_from_env_config() -> Optional[Transport]:
        from openlineage.client.transport.http import HttpTransport, HttpConfig, \
            create_token_provider
        # backwards compatibility: create Transport from
        # OPENLINEAGE_URL and OPENLINEAGE_API_KEY
        if 'OPENLINEAGE_URL' not in os.environ:
            log.error("Did not find openlineage.yml and OPENLINEAGE_URL is not set")
            return
        config = HttpConfig(
            url=os.environ['OPENLINEAGE_URL'],
            auth=create_token_provider({
                "type": "api_key",
                "api_key": os.environ.get('OPENLINEAGE_API_KEY', None)
            })
        )
        return HttpTransport(config)
